#ident "kit/R3KEMU/sbdclock.c: Copyright (c) MIPS Technologies, Inc. All rights reserved."

/*
 * Copyright (c) 1996-2003 MIPS Technologies, Inc.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above
 *      copyright
 *        notice, this list of conditions and the following disclaimer
 *        in the documentation and/or other materials provided with
 *        the distribution.
 *      * Neither the name of MIPS Technologies Inc. nor the names of its
 *        contributors may be used to endorse or promote products derived
 *        from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * R3KEMU/sbdclock.c: timing support functions for R3000 Emulator
 */


#include <stddef.h>
#include <sys/time.h>
#include <sys/clock.h>
#include <mips/cpu.h>
#include <mips/xcpt.h>
#include <assert.h>

#include "kit_impl.h"
#include "sbd.h"

/* emulated count and compare registers */
#define R3KCOUNT	(*(volatile unsigned long *)PA_TO_KVA1(COUNT_ADDR))
#define R3KCOMPARE	(*(volatile unsigned long *)PA_TO_KVA1(COMPARE_ADDR))

static clock_t		(*tickfunc)();
static clock_t		tick;
static clock_t		maxtick;
static unsigned long	cycpertick;
static unsigned long	cycperusec;
static unsigned long	lastcount;

clock_t			_clk_tck;

static int
clkintr (int arg, struct xcptcontext *xcp)
{
    unsigned long nextcount; 
    long cycles, ticks;
    clock_t newtick;

    /* get number of cycles since last tick */
    cycles = R3KCOUNT - lastcount;
    if (cycles < cycpertick)
	cycles = cycpertick;	/* !!! paranoia !!! */

    /* round down to number of whole ticks */
    ticks = cycles / cycpertick;
    cycles -= cycles % cycpertick;

    /* note counter value at last tick */
    lastcount += cycles;

    /* call timer tick function */
    newtick = (*tickfunc) (ticks, xcp);
    
    /* update parameters if they have changed */
    if (newtick != tick) {
	tick = newtick;
	cycpertick = tick * cycperusec;
    }

    /* set next tick point */
    R3KCOMPARE = (nextcount = lastcount + cycpertick);
    
    /* There is a race between reading count and writing compare
       whereby we could set compare to a value "below" the new
       clock value.  Check again to avoid an 80 sec. wait */
    cycles = R3KCOUNT - nextcount;
    while (cycles > 0) {
	R3KCOMPARE = (nextcount += cycpertick);
	cycles = R3KCOUNT - nextcount;
    }

    return 0;
}


void
_sbd_settimer (clock_t newtick)
{
    unsigned long nextcount;
    long cycles;

    assert (newtick <= maxtick);
    if (newtick == tick)
	return;

    tick = newtick;
    cycpertick = tick * cycperusec;

    /* set next tick point */
    R3KCOMPARE = (nextcount = lastcount + cycpertick);
	
    /* make sure we don't wait 80 secs */
    cycles = R3KCOUNT - nextcount;
    while (cycles > 0) {
	R3KCOMPARE = (nextcount += cycpertick);
	cycles = R3KCOUNT - nextcount;
    }
}


clock_t
_sbd_gettimer ()
{
    /* return how many microseconds have passed since last interrupt */
    return (R3KCOUNT - lastcount) / cycperusec;
}


clock_t
_sbd_inittimer (clock_t (*func)())
{
    if (!_clk_tck) {
	/* get CPU frequency */
	_clk_tck = _sbd_pipefreq ();
    }
    
    /* work out timing parameters based on that */
    cycperusec = _clk_tck / 1000000;
    maxtick = 0x7fffffff / cycperusec;

    /* initialise counter */
    R3KCOUNT = (lastcount = 0);
    R3KCOMPARE = 0;

    tickfunc = func;
    intrupt (6, clkintr, 0);
    mips_bissr (SR_IM6);	/* enable interrupt now */

    return maxtick;
}
